/******************************

	CurveLayout.
	
	Like LineLayout except curved.
	
	Copyright 1996, Apple Computer Inc.  
	
	Daniel I. Lipton
	

*******************************/

//#define DRAWHIT
#define DEBUGLEVEL 4

#include "Math.h"
#include "String.h"
#include "GXTypes.h"
#include "GXMath.h"
#include "GXGraphics.h"
#include "GXLayout.h"
#include "GXExceptions.h"
#include "PathWalking.h"

#include "CurveLayout.h"

// GX's shape measurement is not accurate enough, resulting in bad dashing, bad length computation
//		and bad point along curve computations, thus producing crummy kerning for CurveLayout.
//		Use the Joseph/Dan library until GX is fixed.

#define dontUseGXMeasurement

#ifdef	dontUseGXMeasurement
	#include "AccurateShapes.h"
	#define GXPrimitiveShape(target) AccuratePrimitiveShape(target)
	#define GXGetShapeLength(target, index, result) AccurateGetShapeLength(target, index, result)
	#define GXShapeLengthToPoint(target, index, length, point, tangent) AccurateShapeLengthToPoint(target, index, length, point, tangent)
#endif


/************************

	Wrapper for GXGetShapeLength
	which returns a fixed point value.

************************/
Fixed		ClGetFixedShapeLength(gxShape theShape);
Fixed		ClGetFixedShapeLength(gxShape theShape)
	{
		wide			wLength;
		Fixed			pathLength;
				
		GXGetShapeLength(theShape, 0, &wLength);
		pathLength = (Fixed)wLength.lo;
	
		return(pathLength);
	
	}//ClGetFixedShapeLength
	
//-------------------------------------------------------------------------------


/******* 
	Functions pertaining to finding the distance along a
	path closest to a hit point.  Called by ClHitTestCurveLayout
	
	These include:
	
		the primary hit distance testing function
		which uses the shape walker library to find the segment
		which is closest to the hit point and thus the distance
		along the path to that point on that segment
		
		The path walking callback functions which compute the closest
		point on a particular segment to the hit point.
 ********/


/***********************

	Function finds the length along a shape segment
	of the segment's closet point to the hit point.
	
	This is done via a binary search of the distance, 
	 calling GXShapeLengthToPoint to find the point
	 which is closest to the hitPoint.
	 
	 The square of the distance from hitpoint to segment
	 can also be returned if requested.
	 
	Note:
	  While the input parameter to this is a shape object,
	  this only works properly for shapes that are
	  single line or curve segments.
	 
	 The converted length is the result.  
	 The point on the curve can also be returned.
	 The distance (squared) can also be returned.

*************************/

// stop within 1/4 point.

#define searchTolerence 0x00000100
Fixed FindHitLengthAlongSegment(gxShape segment, const gxPoint *hitPoint, gxPoint *pointOnCurve, float *distSquared);
Fixed FindHitLengthAlongSegment(gxShape segment, const gxPoint *hitPoint, gxPoint *pointOnCurve, float *distSquared)
	{
		Fixed				length;
		float				d1, d2, dx, dy;
		gxPoint			startPoint, endPoint;	
		Fixed				startDist, endDist;	
		Boolean			firstTime = true;
		gxShapeType	theType = GXGetShapeType(segment);
		
		/* Sanity check for line or curve segment */
		
		if ( (theType != gxLineType) && (theType != gxCurveType) ) {
		
			GXPostGraphicsError(illegal_type_for_shape);
			return(0);
		
		}//end if
		
		#ifndef dontUseGXMeasurement
			if ( theType == gxCurveType)
				GXSetShapeType(segment, gxPathType);							// GXGetShapeLength crashes on curves.
		#endif

		/* could optimize for lines here using closed formula instead of binary search */
		
		/* Binary search for the distance */
				
		endDist = ClGetFixedShapeLength(segment) - 0x00000008;				// subtract a small amount becuase of round off error in ShapeLengthToPoint
		startDist = ff(0);
		
		GXShapeLengthToPoint(segment, 0, startDist, &startPoint, nil);
		GXShapeLengthToPoint(segment, 0, endDist, &endPoint, nil);

		/* Do a binary search to find the distance into the segment whose point is closest to the hitPoint */
		
		while( (endDist - startDist) > searchTolerence ) {

			dx = FixedToFloat( hitPoint->x - startPoint.x );
			dy = FixedToFloat( hitPoint->y - startPoint.y );
			d1 = dx * dx + dy * dy;
			
			dx = FixedToFloat( hitPoint->x - endPoint.x );
			dy = FixedToFloat( hitPoint->y - endPoint.y );
			d2 = dx * dx + dy * dy;
			
			if (d1 < d2) {
			
				if (distSquared) *distSquared = d1;

				endDist -= ( endDist - startDist) / 2;											// move the end in by half.
				GXShapeLengthToPoint(segment, 0, endDist, &endPoint, nil);	// compute the new point.
						
			} else {

				if (distSquared) *distSquared = d2;

				startDist += ( endDist - startDist) >> 1;												// move the start up by half.
				GXShapeLengthToPoint(segment, 0, startDist, &startPoint, nil);	// compute the new point.
			
			}//end if
			
		}//end while
		
		/* return the actual point on the curve if requested */
		
		if (pointOnCurve != nil) {
		
			pointOnCurve->x = startPoint.x;
			pointOnCurve->y = startPoint.y;	
					
		}//end if
				
		length = startDist;
				
		return(length);
		
	}//FindHitLengthAlongSegment



/* define a structure for pathwalker state used for determining hit length along shape */

typedef struct {

	Boolean			firstSegment;					// is this the first segment being looked at?
	gxPoint			hitPoint;							// the point being tested.
	gxPoint			currentPoint;					// Maintain current point
	gxPoint			firstPointInContour;	// First point in the current contour
	Fixed				totalLength;					// Total accumulated distance so far.
	Fixed				lengthToClosestPoint;	// length into shape of point with shortest distance
	float				shortestDistance;			// shortest distance (squared) to hit point seen so far.
	gxPoint			closestPoint;					// Point on shape for shortest distance seen so far.
	gxShape			segment;							// Make shape object for individual semgnets in here.

} THitLengthRec;


/*******
	Function HitLengthCheckForSmallest:
	
	This function is called from the shape walker segment callbacks
	to see if it is the segment closest to the hit point so far.
	
	Function returns true if the distance from hit to segment was zero. (what are the chances?)
********/
Boolean HitLengthCheckForSmallest(THitLengthRec *hitLength);
Boolean HitLengthCheckForSmallest(THitLengthRec *hitLength)
	{
		Fixed				segmentLength;							// length of this segment.
		gxPoint			closestPointOnSegment;			// the closest point on this segment to hit point
		Fixed				lengthToClosestPoint;				// length along segment to closest point on segment to hit point.
		float				distance;										// distance (squared from hit point to this segment)
	
		segmentLength = ClGetFixedShapeLength(hitLength->segment);		// get the length of this segment.

		lengthToClosestPoint = FindHitLengthAlongSegment(hitLength->segment,
																										 &hitLength->hitPoint, 
																										 &closestPointOnSegment,
																										 &distance);
		
		/* If this is the closest point so far (or the first one), remember it */
		
		if (hitLength->firstSegment || ( distance < hitLength->shortestDistance)) {
		
			hitLength->shortestDistance = distance;

			// do some initialization if necessary.
			if (hitLength->firstSegment) {
				hitLength->totalLength = 0.0;					// Initalize total length if on first segment.
				hitLength->firstSegment = false;			// we're not on first segment any more after this.
			}//end if
			
			// length along shape is total length so far (not including this segment) + lengthToClosest
			hitLength->lengthToClosestPoint = hitLength->totalLength + lengthToClosestPoint;
			hitLength->closestPoint.x = closestPointOnSegment.x;
			hitLength->closestPoint.y = closestPointOnSegment.y;
		
		}//end if
		
		hitLength->totalLength += segmentLength;			// add length of this segment to total length
		
		return( distance == 0.0 ? true : false );
	
	}//HitLengthCheckForSmallest
	
	
	
/***
	 define callbacks for pathwalking.
	 Callbacks will compute the distance from the hit point
	 to thepath's  segment.
	 
	 Search for closest is stopped if we find a zero distance from
	 hit point to shape.
***/ 


Boolean HitLengthLineto(gxPoint *p, THitLengthRec *pWalk);
Boolean HitLengthLineto(gxPoint *p, THitLengthRec *pWalk)
	{
		Boolean			stop;
		gxLine			theSegment;
		
		theSegment.first.x = pWalk->currentPoint.x;
		theSegment.first.y = pWalk->currentPoint.y;
		theSegment.last.x = p->x;
		theSegment.last.y = p->y;
		GXSetLine(pWalk->segment, &theSegment);							// make a shape out of this segment.

		stop = HitLengthCheckForSmallest(pWalk);
				
		pWalk->currentPoint.x = p->x;
		pWalk->currentPoint.y = p->y;
	
		return(stop);
	}

Boolean HitLengthClosepath(THitLengthRec *pWalk);
Boolean HitLengthClosepath(THitLengthRec *pWalk)
	{
		Boolean			stop;
		gxLine			theSegment;
		
		theSegment.first.x = pWalk->currentPoint.x;
		theSegment.first.y = pWalk->currentPoint.y;
		theSegment.last.x = pWalk->firstPointInContour.x;
		theSegment.last.y = pWalk->firstPointInContour.y;
		GXSetLine(pWalk->segment, &theSegment);							// make a shape out of this segment.

		stop = HitLengthCheckForSmallest(pWalk);
				
		pWalk->currentPoint.x = theSegment.last.x;
		pWalk->currentPoint.y = theSegment.last.y;
	
		return(stop);
	}

Boolean HitLengthCurveto(gxPoint p[3], THitLengthRec *pWalk);
Boolean HitLengthCurveto(gxPoint p[3], THitLengthRec *pWalk)
	{
		Boolean			stop = false;
		gxCurve			theSegment;
				
		theSegment.first.x = p[0].x;
		theSegment.first.y = p[0].y;
		theSegment.control.x = p[1].x;
		theSegment.control.y = p[1].y;
		theSegment.last.x = p[2].x;
		theSegment.last.y = p[2].y;
		GXSetCurve(pWalk->segment, &theSegment);									// make a shape out of this segment.
		
		stop = HitLengthCheckForSmallest(pWalk);
				
		pWalk->currentPoint.x = p[2].x;
		pWalk->currentPoint.y = p[2].y;
		
		return(stop);
	}
	
Boolean HitLengthMoveto(gxPoint *p, THitLengthRec *pWalk);
Boolean HitLengthMoveto(gxPoint *p, THitLengthRec *pWalk)
	{
		Boolean		stop = false;

		pWalk->firstPointInContour.x = p->x;
		pWalk->firstPointInContour.y = p->y;
				
		pWalk->currentPoint.x = p->x;
		pWalk->currentPoint.y = p->y;
				
		return(stop);	
	}


/***********************
	Function: ClFindHitLengthAlongPath
	
	Function finds the distance along the path
	of the path's closet point to the hit point.
	
	This is accomplished by finding the closest point on each
	 segment of the path to the hitpoint.  From this, the smallest
	 distanced point is selected.
	 	 
	 The converted distance is the result.  
	 The point on the curve can also be returned.

*************************/

Fixed ClFindHitLengthAlongPath(gxShape path, const gxPoint *hitPoint, gxPoint *pointOnCurve);
Fixed ClFindHitLengthAlongPath(gxShape path, const gxPoint *hitPoint, gxPoint *pointOnCurve)
	{
		THitLengthRec			walker;						// context for walking the segments with ShapeWalker.
		Boolean						stoppedEarly;

		// Initialize the walker.
		walker.firstSegment = true;
		walker.hitPoint.x = hitPoint->x;
		walker.hitPoint.y = hitPoint->y;
		walker.segment = GXNewShape(gxEmptyType);

		/* Invoke shape walker to find the hit distance and point */
		stoppedEarly = ShapeWalker(path, HitLengthMoveto, HitLengthLineto, 
																		 HitLengthCurveto, HitLengthClosepath, &walker);

		GXDisposeShape(walker.segment);
		
		/* return the actual point on the curve if requested */
		
		if (pointOnCurve != nil) {
		
			pointOnCurve->x = walker.closestPoint.x;
			pointOnCurve->y =  walker.closestPoint.y;	
					
		}//end if
		
		#ifdef DRAWHIT
		{
			gxRectangle			aRect;
			aRect.top = walker.closestPoint.y - ff(2);
			aRect.left = walker.closestPoint.x - ff(2);
			aRect.bottom = aRect.top + ff(5);
			aRect.right = aRect.left + ff(5);
			GXDrawRectangle(&aRect, gxEvenOddFill);
		}
		#endif
						
		return(walker.lengthToClosestPoint);
		
	}//ClFindHitLengthAlongPath

/****** end of functions for hit testing *******/

//-------------------------------------------------------------------------------


/*******

	Functions for positioning glyphs along the path
	
*******/

/***********************
	
	Function applys the path to a glyph shape so the
	glyphs flow along the path shape. Hence: CurveLayout
	
	The point on the curve we want the tangent for is that of the horizontal
	center of the glyph's bounding box.  The point corresponding to the horizontal
	bounding box will be considered the origin of rotation for the glyph.  If we
	define a vector V that is the glyph's origin minus this horizontal bounding box center,
	we can rotate that vector V about the point on the curve to determine the new glyph 
	position.  This would be easier to describe with a picture.
	
	Function returns true if all of the glyph's didn't fit on the path.
	
************************/
Boolean ApplyPathToGlyphs(gxShape thePath, gxShape theGlyphs);
Boolean ApplyPathToGlyphs(gxShape thePath, gxShape theGlyphs)
	{
		long					glyphCount;
		gxRectangle*	glyphBoxes;
		gxPoint*			origins;
		gxPoint*			tangents;
		gxPoint				relativePosition;
		long*					advances;
		gxPoint				newPoint, oldTangent, newTangent;
		gxMapping			aMapping;
		Ptr						workspace;
		long					idx;
		gxPoint				pointToMapOnCurve;
		gxShapeType		theType;
		wide					wLength;
		Fixed					shapeLength;	
		Boolean				tooBig = false;	
		
		// some sanity checking.
		theType = GXGetShapeType(theGlyphs);
		if (theType == gxEmptyType)
			return(false);
		else if (theType != gxGlyphType) {
			GXPostGraphicsError(illegal_type_for_shape);
			return(false);
		}//end if
		
		// Allocate work space memory.
		
		(void)GXGetGlyphs(theGlyphs, &glyphCount, nil, nil, nil, nil, nil, nil, nil);
		
		workspace = NewPtr(glyphCount * sizeof(gxRectangle) + 			// bounding boxes
											(glyphCount+1) * sizeof(gxPoint) +				// positions.
											glyphCount * sizeof(gxPoint) +						// tangents.
											4 * ( (glyphCount + 31) / 32 ) ) ;				// advances.
											
		require_action(workspace, failed_allocate, GXPostGraphicsError(out_of_memory););	
	
		origins = (gxPoint*)workspace;
		glyphBoxes = (gxRectangle*)(origins + glyphCount + 1);
		tangents = (gxPoint*)(glyphBoxes + glyphCount);
		advances = (long*)(tangents + glyphCount);
		
		// Get the data from the shape.
		GXGetGlyphs(theGlyphs, nil, nil, nil, nil, tangents, nil, nil, nil);
		GXGetGlyphMetrics(theGlyphs, origins, glyphBoxes, nil);

		// we don't want to exceed the path length while dashing.
		GXGetShapeLength(thePath, 0, &wLength);	
		shapeLength = (Fixed)wLength.lo;
		
		for (idx = 0; idx < glyphCount; ++idx) {
		
			// Compute the horizontal center of the glyph.  This is the point to map along path.
			pointToMapOnCurve.x = glyphBoxes[idx].left + (glyphBoxes[idx].right - glyphBoxes[idx].left) / 2;
			pointToMapOnCurve.y = 0;
			
			// If the right side of the glyph goes off the end of the curve, then stop.
			if (glyphBoxes[idx].right > shapeLength)
				break;
			
			// Compute the position of the glyph relative to the bounding box horizontal center.
			relativePosition.x = origins[idx].x - pointToMapOnCurve.x;
			relativePosition.y = origins[idx].y - pointToMapOnCurve.y;
			
			// Now find the new location and tangent for where the horizontal bounding box center should be
			//  along the curve.
			GXShapeLengthToPoint(thePath, 0, pointToMapOnCurve.x, &newPoint, &newTangent);
			
			// The new position will be the glyph's position relative to the horizontal bbox center
			//	 rotated by the tangent.  First rotate.
			ResetMapping(&aMapping);
			aMapping.map[0][0] =  newTangent.x;
			aMapping.map[1][0] =  -newTangent.y;
			aMapping.map[0][1] =  newTangent.y;
			aMapping.map[1][1] =  newTangent.x;
			MapPoints(&aMapping, 1, &relativePosition);
			
			// Now position this relative to the new point.
			origins[idx].x = newPoint.x + relativePosition.x;
			origins[idx].y = newPoint.y + relativePosition.y;
			
			
			// Concatonate the new tangent with the old.
			oldTangent.x = tangents[idx].x;
			oldTangent.y = tangents[idx].y;
			tangents[idx].x = FixedMultiply( oldTangent.x, newTangent.x ) - FixedMultiply( oldTangent.y, newTangent.y);
			tangents[idx].y = FixedMultiply( oldTangent.x, newTangent.y ) + FixedMultiply( oldTangent.y, newTangent.x);
			
		}//end for

		// Handle case where we had to cut glyphs cause path isn't long enough.
		if (idx < glyphCount) {
			glyphCount = idx;
			GXSetGlyphs(theGlyphs, glyphCount, nil, nil, nil, nil, nil, nil);			// set the new length.
			tooBig = true;
		}//end if
		
		// Now set the glyph shape for new positions, tangents, advances.
		/* Make all of the advance bits 1 for absolute positioning of each glyph */
		for (idx = 0; idx < ( (glyphCount + 31) / 32 ); ++idx)
			advances[idx] = 0xFFFFFFFF;
		
		GXSetGlyphPositions(theGlyphs, 1, glyphCount, advances, origins);
		GXSetGlyphTangents(theGlyphs, 1, glyphCount, tangents);
				
		DisposePtr(workspace);
		
failed_allocate:
		return(tooBig);
	
	}//ApplyPathToGlyphs
	
	

/**********************

	Routine to make a glyph shape
	that represents the layout shape 
	drawn on the path.
	
	theLayout:			a shape of type text, glyph, or layout.
	thePath:				path to draw it on.
	old:						If you have an old one, pass it in to avoid creating new one.
	tooBig:					returned:  text was too long for the path.
	
	returned:				the glyph shape that is layout along path.
		
***********************/

gxShape MakeTextOnPath(gxShape theLayout, gxShape thePath, gxShape old, Boolean *tooBig);
gxShape MakeTextOnPath(gxShape theLayout, gxShape thePath, gxShape old, Boolean *tooBig)
	{
		gxShape		theGlyphs = GXCopyToShape(old, theLayout);
				
		GXPrimitiveShape(theGlyphs);
	
		#ifdef USEDASHINGFORGLYPHS
		{
			gxDashRecord			theDash;
			gxShape						pathCopy = GXCopyToShape(nil, thePath);
			
			theDash.dash = theGlyphs;
			theDash.advance = 0;
			theDash.phase = 0;
			theDash.scale = ff(1);
			theDash.attributes = gxBreakDash;
			
			GXSetShapeDash(pathCopy, &theDash);
			GXSetShapePen(pathCopy, ff(1));

			GXPrimitiveShape(pathCopy);
			*tooBig = false;							// difficult to implement on top of dashing.
			
			GXDisposeShape(theGlyphs);
			return(pathCopy);
		}
		#else
			// Dashing of glyph shapes is really busted.
			//  Besides, in our own code we can exactly control the algorithm.
			*tooBig = ApplyPathToGlyphs(thePath, theGlyphs);
		#endif

		return(theGlyphs);
	
	}//MakeTextOnPath



/*** End of functions for positioning text on the path ***/


//-------------------------------------------------------------------------------

/*******************************

	Wrapper for GXGetLayoutHighlight
	for layout on path.  This is implemented
	by getting the layout highlight for each
	character specified seperately.  These are made
	into a multi-contour polygon (one for each character)
	which is then used as a dash on the path.  Calling 
	GXPrimitiveShape on such a path will result in a multi-contour
	polygon which will have the appropriate characters highlighted.

*********************************/
gxShape GetLayoutHighlightOnPath(gxShape theLayout, gxShape thePath, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight);
gxShape GetLayoutHighlightOnPath(gxShape theLayout, gxShape thePath, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight)
	{
		gxShape				theShape;
		gxShape				newHighlight;
		gxShape				oneCharHighlight = nil;
		gxByteOffset	idx;
		gxDashRecord	theDash;
		
		newHighlight = GXNewShape(gxEmptyType);
		
		for (idx = startOffset; idx < endOffset; ++idx) {
		
			oneCharHighlight = GXGetLayoutHighlight(theLayout, idx, idx+1, hilightType, oneCharHighlight);

			if (oneCharHighlight != nil) {
				
				/* Add the highlight for the individual character to the highlight shape */
				
				GXSetShapeParts(newHighlight, 0, 0, oneCharHighlight, gxBreakLeftEdit);
									
			}//end if//end if			
		
		}//end for
		
		GXDisposeShape(oneCharHighlight);
		
		theShape = GXCopyToShape(highlight, thePath);
		GXSetShapePen(theShape, ff(1));						// we don't want dashes to scale.
		GXSetShapeFill(theShape, gxFrameFill);		
		theDash.attributes = gxBreakDash;
		theDash.dash = newHighlight;
		theDash.advance = 0;									// we want only one repeat
		theDash.phase = ff(0);
		theDash.scale = ff(1);
		GXSetShapeDash(theShape, &theDash);
		GXDisposeShape(newHighlight);
		GXPrimitiveShape(theShape);
		GXSetShapeDash(theShape, nil);


		/*****
			Change the fill of the result to winding.
			Since when bent around, contours can overlap
			we want the overlaps to fill, too.  With Even-odd,
			they wouldn't be filled and the highlihgt would look
			funny.
		*****/
		
		GXSetShapeFill(theShape, gxWindingFill);
		
		return(theShape);
		
	}//GetLayoutHighlightOnPath 



/************************** Following routines deal with contigus highlighting *******************/

/****************************

	Routine: GetLinesFromContour.
	
	This routine computes the two horizontal segments
	from a contour of the highlight shape
	
	contour:				(input) 	Point to the contour of the shape.
	bottomSegment:	(output)	The computed bottom line segment of the contour.
	topSegment		:	(output)	The computed top line segment of the contour.
	
******************************/
void GetLinesFromCountour(Ptr contour, gxLine *bottomSegment, gxLine *topSegment);
void GetLinesFromCountour(Ptr contour, gxLine *bottomSegment, gxLine *topSegment)
	{
		gxPoint	p1, p2, p3, p4;
		long		nPoints = *(long*)contour;
		gxLine	seg1, seg2;
		Fixed		swapper;
		
		contour += sizeof(long);
		
		/* read the points out of the contour */
		
		p1.x = ((gxPoint*)contour)->x;
		p1.y = ((gxPoint*)contour)->y;
		contour += sizeof(gxPoint);
		
		p2.x = ((gxPoint*)contour)->x;
		p2.y = ((gxPoint*)contour)->y;
		contour += sizeof(gxPoint);
		
		p3.x = ((gxPoint*)contour)->x;
		p3.y = ((gxPoint*)contour)->y;
		contour += sizeof(gxPoint);

		p4.x = ((gxPoint*)contour)->x;
		p4.y = ((gxPoint*)contour)->y;
		
		/* now figure out how to make line segments */
		
		if ( nPoints == 4 ) {
		
			/* for 4 points, test each set of pairs to find the horizontal segments */
			
			if ( (p1.y == p2.y) && (p3.y == p4.y) ) {
			
				seg1.first.y = seg1.last.y = p1.y;
				seg1.first.x = p1.x;
				seg1.last.x = p2.x;
				
				seg2.first.y = seg2.last.y = p3.y;
				seg2.first.x = p3.x;
				seg2.last.x = p4.x;
			
			} else if ((p1.y == p3.y) && (p2.y == p4.y) ) {
			
				seg1.first.y = seg1.last.y = p1.y;
				seg1.first.x = p1.x;
				seg1.first.y = p3.x;
				
				seg2.first.y = seg2.last.y = p2.y;
				seg2.first.x = p2.x;
				seg2.last.x = p4.x;
			
			} else if ( (p1.y == p4.y) && ( p2.y == p3.y) ) {

				seg1.first.y = seg1.last.y = p1.y;
				seg1.first.x = p1.x;
				seg1.last.x = p4.x;
				
				seg2.first.y = seg2.last.y = p2.y;
				seg2.first.x = p2.x;
				seg2.last.x = p3.x;
				
			} else {
			
				GXPostGraphicsError(inconsistent_parameters);	
			
			}//end if
			
		} else if ( nPoints == 3) {
		
			DebugStr("\pNeed to handle the 3 point case");
		
		} else {
			
			GXPostGraphicsError(inconsistent_parameters);	
			
		}//end if
		
		/* now make sure both segments go from left to right */
		
		if (seg1.first.x > seg1.last.x) {
		
			swapper = seg1.first.x;
			seg1.first.x = seg1.last.x;
			seg1.last.x = swapper;
		
		}//end if

		if (seg2.first.x > seg2.last.x) {
		
			swapper = seg2.first.x;
			seg2.first.x = seg2.last.x;
			seg2.last.x = swapper;
		
		}//end if

		/* now make the one smaller y the upper and the other one the lower */
		
		if ( seg1.first.y < seg2.first.y ) {
		
			memcpy(topSegment, &seg1, sizeof(gxLine));
			memcpy(bottomSegment, &seg2, sizeof(gxLine));
		
		} else {

			memcpy(topSegment, &seg2, sizeof(gxLine));
			memcpy(bottomSegment, &seg1, sizeof(gxLine));
		
		}//end if		
		
	}//GetLinesFromCountour
	

/*******************************

	Routine combines the two shapes into
	a filled highlight region.  The upper 
	shape represents the top segments
	of the highlight returned from layout (after
	put through dashing contortions) 	and the lower
	shape the lower segments.
	
	Assume that upper and lower start with same
	number of contours.

********************************/
gxShape CombineUpperAndLower(gxShape upper, gxShape lower, gxShape old);
gxShape CombineUpperAndLower(gxShape upper, gxShape lower, gxShape old)
	{
		gxShape			result;
		long				numContours;
		Ptr					upperWalker, lowerWalker;
		long				upperPointCount, lowerPointCount;
		long				insertIndex, idx;
		gxPaths			*upperPathsPtr, *lowerPathsPtr;
				
		/* reverse the contours of the upper path because they are in the same direction as lower */
		
		numContours = GXCountShapeContours(upper);
		for (idx = 1; idx <= numContours; ++idx)
			GXReverseShape(upper, idx);

		GXLockShape(upper);
		upperWalker = (Ptr)GXGetShapeStructure(upper, nil);
		GXLockShape(lower);
		lowerWalker = (Ptr)GXGetShapeStructure(lower, nil);
				
		check(numContours == *(long*)lowerWalker);
		
		upperWalker += sizeof(long);			// skip past number of contours
		lowerWalker += sizeof(long);			//   for both walkers.
		insertIndex = 0;

		/* eather make a new shape or a copy a new shape into the old reference */
		
		if (old == nil) {
			
			result = GXNewShape(gxPathType);
			
		} else {
		
			/* this is so we emulate layout's API feature of reusing the same shape reference
					even though we are always creating a new shape */
					
			gxShape temp = GXNewShape(gxPathType);
			result = GXCopyToShape(old, temp);
			GXDisposeShape(temp);
		
		}//end if
				
		/* Make a new shape such that each contour is a combination of a lower contour and an upper contour */

		for (idx = 0; idx < numContours; ++idx) {
		
			upperPointCount = *(long*)upperWalker;
			lowerPointCount = *(long*)lowerWalker;
			
			/****
				we need to pass a paths structure into GXSetPathParts, so point behind each contour to 
					make it look like a paths And put in 1 for 1 contour
			***/
			upperPathsPtr = (gxPaths*)(upperWalker - sizeof(long) );
			lowerPathsPtr = (gxPaths*)(lowerWalker - sizeof(long) );
			
			upperPathsPtr->contours = 1;
			lowerPathsPtr->contours = 1;
			
			GXSetPathParts(result, 0, 0, lowerPathsPtr, gxBreakLeftEdit);
			GXSetPathParts(result, 0, 0, upperPathsPtr, gxBreakNeitherEdit);
			
			/* update the walkers for reading the next contour */
			
			upperWalker += sizeof(long) +												// skip point count
										 upperPointCount * sizeof(gxPoint) +	// and the points
										 4 * ( (upperPointCount + 31)/32 );		// and the control bits.

			lowerWalker += sizeof(long) +												// skip point count
										 lowerPointCount * sizeof(gxPoint) +	// and the points
										 4 * ( (lowerPointCount + 31)/32 );		// and the control bits.

		}//end for
				
		GXSetShapeFill(result, gxWindingFill);
		
		GXUnlockShape(upper);
		GXUnlockShape(lower);

		return(result);
			
	}//CombineUpperAndLower

/************************************

	Routine: ContiguousGetLayoutHighlightOnPath
	
	This routine builds a shape that is the highlight
	bent along the path.  It is contiguous as oppsed
	to broken on each character.  This is done using
	the bend dashing feature.  
	
	bend dashing is a way to find the curved segments
	of a particular length of a path.  the upper and lower
	line segments are used seperately to bend-dash the path.
	the resulting upper and lower curved segments are then combined
	into filled shapes.  The upper one is outset by ascent of layout
	and inner is inset by decent of layout.
	
	Method implemented in this code is Patent Pending. 
			(Apple Computer, Inc. Daniel I. Lipton)

*************************************/
gxShape ContiguousGetLayoutHighlightOnPath(gxShape theLayout, gxShape thePath, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight);
gxShape ContiguousGetLayoutHighlightOnPath(gxShape theLayout, gxShape thePath, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight)
	{
		gxShape				realHighlight;												// the highlight returned by layout to operate upon.
		long					numContours, nPoints;
		Ptr						polyWalker;														// pointer used for walking poly data.
		gxShape				upperSegments, lowerSegments;					// make shapes that represent upper and lower of highlight
		gxShape				upperPath, lowerPath;									// This will be the versions that are curved.
		gxShape				aLine;																// a shape used for building upper and lower segments.
		gxLine				topSegment, bottomSegment;						// used for constructing upper and lower segments.
		gxDashRecord	theDash;
		gxShape				result;
		Fixed					lineAscent, lineDecent;
		Fixed					pathLength;
		long					idx;

		pathLength = ClGetFixedShapeLength(thePath);
					
		/* find the uncurved highlight */
		
		realHighlight = GXGetLayoutHighlight(theLayout, startOffset, endOffset, hilightType, nil);

		/* Walk the highlight returned from layout to find upper and lower segments */

		GXLockShape(realHighlight);	
		polyWalker = (Ptr)GXGetShapeStructure(realHighlight, nil);
		
		aLine = GXNewShape(gxLineType);
		
		numContours = *(long*)polyWalker;			// first long is number of contours.
		polyWalker += sizeof(long);
		
		upperSegments = GXNewShape(gxPolygonType);
		GXSetShapeFill(upperSegments, gxFrameFill);
		lowerSegments = GXNewShape(gxPolygonType);
		GXSetShapeFill(lowerSegments, gxFrameFill);
		
		/* Get the upper and lower segments of each contour of the highlight shape */
		
		for (idx = 0; idx < numContours; ++idx) {
	
			/* first field in contour is number of points */
			
			nPoints = *(long*)polyWalker;
			
			GetLinesFromCountour(polyWalker, &bottomSegment, &topSegment);
			
			polyWalker += sizeof(long) + nPoints * sizeof(gxPoint);

			/* make sure points won't wrap around when dashed */

			if (bottomSegment.first.x < 0)
				bottomSegment.first.x = 0;
			else if (bottomSegment.first.x > pathLength)
				bottomSegment.first.x = pathLength;
		
			if (bottomSegment.last.x < 0)
				bottomSegment.last.x = 0;
			else if (bottomSegment.last.x > pathLength)
				bottomSegment.last.x = pathLength;
								
			if (topSegment.first.x < 0)
				topSegment.first.x = 0;
			else if (topSegment.first.x > pathLength)
				topSegment.first.x = pathLength;
							
			if (topSegment.last.x < 0)
				topSegment.last.x = 0;
			else if (topSegment.last.x > pathLength)
				topSegment.last.x = pathLength;
								
			/* Add the segments to the upper and lower shapes, we are only interested in X coordinates. */

			GXSetLine(aLine, &bottomSegment);			
			GXSetShapeParts(lowerSegments, 0, 0, aLine, gxBreakLeftEdit);
						
			GXSetLine(aLine, &topSegment);
			GXSetShapeParts(upperSegments, 0, 0, aLine, gxBreakLeftEdit);
		
		}//end for

		GXUnlockShape(realHighlight);
		GXDisposeShape(realHighlight);				// we're done with it now.
		
		GXDisposeShape(aLine);

		/* now use hairline dashing to make highlight segments follow path */
				
		upperPath = GXCopyToShape(nil, thePath);
		GXSetShapePen(upperPath, ff(0));			// we don't want dashes to scale.
		theDash.dash = upperSegments;
		theDash.attributes = gxBendDash;			// make dashes bend with path.
		theDash.advance = 0;									// we want only one repeat
		theDash.phase = ff(0);
		theDash.scale = ff(1);
		GXSetShapeDash(upperPath, &theDash);
		GXDisposeShape(upperSegments);
		GXPrimitiveShape(upperPath);
		GXSetShapeDash(upperPath, nil);

		lowerPath = GXCopyToShape(nil, thePath);
		GXSetShapePen(lowerPath, ff(0));			// we don't want dashes to scale.
		theDash.dash = lowerSegments;
		theDash.attributes = gxBendDash;
		theDash.advance = 0;									// we want only one repeat
		theDash.phase = ff(0);
		theDash.scale = ff(1);
		GXSetShapeDash(lowerPath, &theDash);
		GXDisposeShape(lowerSegments);
		GXPrimitiveShape(lowerPath);
		GXSetShapeDash(lowerPath, nil);

		/* Now combine the upper and lower paths into a filled shape for the highlight */
		
		GXGetLayoutSpan(theLayout, &lineAscent, &lineDecent);
		
		GXInsetShape(lowerPath, lineDecent);
		GXInsetShape(upperPath, -lineAscent);
		
		GXSetShapeType(upperPath, gxPathType);		// CombineUpperandLower only works on paths.
		GXSetShapeType(lowerPath, gxPathType);
		
		result = CombineUpperAndLower(upperPath, lowerPath, highlight);
			
		GXDisposeShape(upperPath);
		GXDisposeShape(lowerPath);

		return(result);
	
	}//ContiguousGetLayoutHighlightOnPath 


/************************** End of routines for contiguous highlightint ********************/

//-------------------------------------------------------------------------------



/********************

	GetLayoutCaretOnPath:
	 Returns the caret shape for a layout on a path.
	 
	Use the same dashing tricks.  Get the caret,
	use it as a dash shape.
	
*********************/
gxShape GetLayoutCaretOnPath(gxShape layout, gxShape thePath, gxByteOffset offset, gxHighlightType highlightType, gxCaretType caretType, gxShape caret);
gxShape GetLayoutCaretOnPath(gxShape layout, gxShape thePath, gxByteOffset offset, gxHighlightType highlightType, gxCaretType caretType, gxShape caret)
	{
		gxShape				result;
		gxShape				realCaret;			// caret not on path.
		gxDashRecord	theDash;
		gxShapeFill		theFill = GXGetShapeFill(thePath);
				
		realCaret = GXGetLayoutCaret(layout, offset, highlightType, caretType, nil);		

		result = GXCopyToShape(caret, thePath);													// copy into caret passed in (if it is nil, right thing happens).
		GXSetShapePen(result, ff(1));																		// we don't want dashes to scale.
		theDash.attributes = gxBreakDash;
		theDash.dash = realCaret;
		theDash.advance = 0;																						// we want only one repeat
		theDash.phase = ff(0);
		theDash.scale = ff(1);
		GXSetShapeDash(result, &theDash);

		/* make sure we have a framed shape for the path */
		
		if ( (theFill != gxFrameFill) && (theFill != gxClosedFrameFill) )
			GXSetShapeFill(result, gxClosedFrameFill);

		/* Now apply the dash to get the caret on the path */
		
		GXPrimitiveShape(result);
		GXSetShapeDash(result, nil);			// we don't want the result redashed when drawn.
		
		if ( GXGetShapeType(result) == gxEmptyType ) {

			gxPoint			position, tangent;
			Fixed				distance;
			gxMapping		aMapping;
								
			/****
				Caret could have been off the curve because edges of characters fall over.
				Dashing will result in an empty shape in this case.
				Make best guess based on tangent of curve at end point.
				This means caret may not be where Layout would've really
				wanted it, but it's the best we can do.
			****/
		
			if (offset == 0) {
			
				GXShapeLengthToPoint(thePath, 0, ff(0), &position, &tangent);
			
			} else {

				distance = ClGetFixedShapeLength(thePath);
				GXShapeLengthToPoint(thePath, 0, distance, &position, &tangent);
						
			}//end if
			
			result = GXCopyToShape(result, realCaret);
			ResetMapping(&aMapping);
			aMapping.map[0][0] =  tangent.x;
			aMapping.map[1][0] = -tangent.y;
			aMapping.map[0][1] =  tangent.y;
			aMapping.map[1][1] =  tangent.x;
			GXMapShape(result, &aMapping);
			GXMoveShape(result, position.x, position.y);
					
		}//end if

		GXDisposeShape(realCaret);

		return(result);
		
	}//GetLayoutCaretOnPath

/*******************************

	Wrapper for GXHitTestLayout
	for layout on path.  
	
	The distance of the point into the 
	path is computed and then that is
	passed into GXHitTestLayout to
	compute the hit info.

*********************************/
gxByteOffset HitTestLayoutOnPath(gxTransform clTransform, gxShape layout, gxShape path, const gxPoint *hitDown, gxHighlightType highlightType, gxLayoutHitInfo *hitInfo, gxShape hitTrackingArea);
gxByteOffset HitTestLayoutOnPath(gxTransform clTransform, gxShape layout, gxShape path, const gxPoint *hitDown, gxHighlightType highlightType, gxLayoutHitInfo *hitInfo, gxShape hitTrackingArea)
	{
		gxByteOffset			result = 0;
		Fixed							lineAscent, lineDecent;
		gxShape						pathToTest;
		gxStyleAttribute	attributes;
		gxHitTestInfo			resultInfo;
		gxShapePart				resultMask;
		Fixed							distance;
		gxTransform				hitTransform;
		gxPoint						testPoint;
		gxMapping					theMapping, iMapping;
				
		/* Make a copy of the path that can be outside-framed by the ascent+decent for hit testing */
		
		GXGetLayoutSpan(layout, &lineAscent, &lineDecent);
								
		pathToTest = GXCopyToShape(nil, path);

		GXInsetShape(pathToTest, lineDecent);									// Inset by the decent.
		GXSetShapePen(pathToTest, lineDecent + lineAscent);		// outside frame by ascent+decent.
		attributes = GXGetShapeStyleAttributes(pathToTest);
		attributes &= ~gxInsideFrameStyle;
		attributes += gxOutsideFrameStyle;
		GXSetShapeStyleAttributes(pathToTest, attributes);

		/* run the hit point through inverse of curveLayout's mapping so hit testing is in identity space */
		
		GXGetTransformMapping(clTransform, &theMapping);
		InvertMapping(&iMapping, &theMapping);
		testPoint.x = hitDown->x;
		testPoint.y = hitDown->y;
		MapPoints(&iMapping, 1, &testPoint);
				
		/* Now hit test the thing */
		
		hitTransform = GXNewTransform();										// curve should have pure transform for hit testing.  We will map points ourselves.
		GXSetTransformHitTest(hitTransform, gxPenPart, ff(0));
		
		GXSetShapeTransform(pathToTest, hitTransform);
		resultMask = GXHitTestShape(pathToTest, &testPoint, &resultInfo);
		GXDisposeShape(pathToTest);
		
		if (resultMask) {
		
			gxPoint			layoutHit;
									
			/* Find the distance along the path whose point is closest to the hit point */

			distance = ClFindHitLengthAlongPath(path, &testPoint, nil);

			layoutHit.x = distance;
			layoutHit.y = ff(0);
			
			/* Use this distance as an offset in the X direction from the layout's location to hit test the layout */

			result = GXHitTestLayout(layout, &layoutHit, highlightType, hitInfo, hitTrackingArea);
			
			/* Lay the hit tracking area along the path */

			if (hitTrackingArea) {
			
				DebugStr("\pNot handling hit tracking area properly yet");
			
			}//end if
												
		}//end if
		
		GXDisposeTransform(hitTransform);
		
		return(result);
		
	}//HitTestLayoutOnPath



/*********************************** Structure of Private Cache ***********************************************/
	

typedef struct {

	long				flags;
	gxShape			theLayout;					// layout shape to bind to path.
	gxShape			thePath;						// path to bind to.
	
} curveLayoutRec;

#define curveLayoutTag 'crvL'

/*********************************************** exported APIs ***********************************************/


/***********************************

	Routine:	ClGetCurveLayout
	
	Routine can be called to retrieve the original
	shapes that were used to make the text on the path.
	
	
	return result:  curve layout flags.
	
************************************/
long ClGetCurveLayout(gxShape input, gxShape *theText, gxShape *thePath)
	{
		gxTag							theTag;
		curveLayoutRec		*theData;
		long							result = 0;
	
		if ( GXGetShapeTags(input, curveLayoutTag, 1, 1, &theTag) ) {
		
			GXLockTag(theTag);
			theData = (curveLayoutRec*)GXGetTagStructure(theTag, nil);
			result = theData->flags;
			
			if (!(result & clSelfContained)) {
			
				if (theText)
					*theText = theData->theLayout;
				
				if (thePath)
					*thePath = theData->thePath;
			
			} else {
			
				if (theText)
					GXGetPictureParts(input, 2, 1, theText, nil, nil, nil);
					
				if (thePath)
					GXGetPictureParts(input, 3, 1, thePath, nil, nil, nil);
			
			}//end if
								
			GXUnlockTag(theTag);
														
		} else {
				
			/* If the shape didn't have our private tag, then it wasn't curveLayout */
			
			GXPostGraphicsError(illegal_type_for_shape);
		
		}//end if
		
		return(result);
	
	}//ClGetCurveLayout
	
	
/***********

	Routine: ClNewCurveLayout:
	
	Routine makes a new shape that is the association
	of the text and the path.  The actual shape type
	is a glyph shape which can be drawn.  Hence, for drawing
	the client can call GXDrawShape normally.
	
	Other manipulations should go through the API.
	
	returns a curve layout shape.
	
	Transforms of original shapes ignored.  Client should
	manipulate transform of this shape for all operations.
	The original shapes will have their transforms replaced.
	The calling application should not assume they will 
	stay the same.
	
************/
gxShape	ClNewCurveLayout(gxShape theText, gxShape thePath, Boolean selfContained)
	{
		gxShape						result;
		curveLayoutRec		tagData;
		gxTag							theTag;
		Boolean						tooBig;
		gxShapeAttribute	attributes;
		gxPoint						position;
		gxTransform				clearTransform = GXNewTransform();
				
		if ( GXGetShapeType(theText) != gxLayoutType ) {
		
			GXPostGraphicsError(illegal_type_for_shape);
			return(nil);
		
		}//end if
				
		/* create a new empty shape and tag it with the input shapes */
		
		if (!selfContained) {
			/* store the original shapes in the tag */
			tagData.theLayout = GXCloneShape(theText);
			tagData.thePath = GXCloneShape(thePath);
		}//end if
		
		tagData.flags = clNoFlags;
		if (selfContained)
			tagData.flags |= clSelfContained;
		
		/* Make sure the layout shape has an identity transform */
		
		GXSetShapeTransform(theText, clearTransform);
		GXDisposeTransform(clearTransform);
		
		/* Also Rest the layout's Y position to be zero, we always want layout to hug curve, but x value will determine indent */
		
		GXGetLayout(theText, nil, nil, nil, nil, nil, nil, nil, nil, &position);
		position.y = ff(0);
		GXSetLayout(theText, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, &position);

		/* doesn't make sense to grid metrics on a curve */
		
		attributes = GXGetShapeAttributes(theText)  | gxNoMetricsGridShape;	// hinted metrics will screw us up when rotated.
		GXSetShapeAttributes(theText, attributes);  

		result = MakeTextOnPath(theText, thePath, nil, &tooBig);		// pass nil for old to create a new one every time.

		if (tooBig)
			tagData.flags |= clTextTooBigForPath;
			
		if (!(tagData.flags & clSelfContained)) {
		
			/* Give the result a default transform */
			
			GXSetShapeTransform(result, GXGetShapeTransform( GXGetDefaultShape(gxGlyphType)) );
			
			attributes = GXGetShapeAttributes(result);
			attributes |= gxMapTransformShape;
			GXSetShapeAttributes(result, attributes);
					
			/* Hide the cache in the shape with a tag */
			
			theTag =  GXNewTag(curveLayoutTag, sizeof(curveLayoutRec), &tagData);
			
			GXSetShapeTags(result, curveLayoutTag, 1, 0, 1, &theTag);
			
			GXDisposeTag(theTag);
	
			return(result);
			
		} else {
			/* Make the curve layout object a picture shape that contains the glyph shape
					to draw as well as the original shapes passed in. */
					
			gxShape					pictResult;
			gxTransform			dontDrawTransform = GXNewTransform();
			gxShape					emptyClip = GXNewShape(gxEmptyType);
			
			GXSetTransformClip(dontDrawTransform, emptyClip);		// this transform will result in no drawing.
			pictResult = GXNewShape(gxPictureType);							// create the picture for the curveLayout.
			
			theTag =  GXNewTag(curveLayoutTag, sizeof(curveLayoutRec), &tagData);
			
			GXSetShapeTags(pictResult, curveLayoutTag, 1, 0, 1, &theTag);
			
			GXDisposeTag(theTag);
			
			// The first two entries in the picture are the original text and path.
			//  They are given an overriding transform that causes no drawing.
			// The third entry in the picture is the glyph shape we want to draw.
			GXSetPictureParts(pictResult, 1, 0, 1, &result, nil, nil, nil);
			GXSetPictureParts(pictResult, 2, 0, 1, &theText, nil, nil, &dontDrawTransform);
			GXSetPictureParts(pictResult, 3, 0, 1, &thePath, nil, nil, &dontDrawTransform);

			GXDisposeShape(result);		// picture owns it now.
			GXDisposeTransform(dontDrawTransform);
			GXDisposeShape(emptyClip);
			
			return(pictResult);

		}//end if
		
	}//ClNewCurveLayout

#include "Stdio.h"
/*******************************

	Routine: ClDisposeCurveLayout.
	
	Routine is called by client instead of GXDisposeShape
	for shapes created by calling ClNewCurveLayout.
	
	Note if the client created a self contained curve layout
	it is sufficient to call GXDisposeShape instead.
	
*********************************/
void  ClDisposeCurveLayout(gxShape curveLayout)
	{
		long		flags = ClGetCurveLayout(curveLayout, nil, nil);
		
		if (!(flags & clSelfContained)){
			gxShape		text, path;
			
			ClGetCurveLayout(curveLayout, &text, &path);
			
			GXDisposeShape(text);
			GXDisposeShape(path);
			
		}//end if
		
		GXDisposeShape(curveLayout);
		
	}//ClDisposeCurveLayout


/**********************************

	Routine ClChangedCurveLayout
	
	This routine must be called by the client
	whenever the text or the path are modified.
	

************************************/
void ClChangedCurveLayout(gxShape target)
	{

		gxTag							theTag;
		curveLayoutRec		*theData;
		Boolean						tooBig;
		gxTransform				preserveTransform;
		gxPoint						position;
		gxShape						theLayout, thePath, theGlyph;
	
		if ( GXGetShapeTags(target, curveLayoutTag, 1, 1, &theTag) ) {
		
			GXCloneTag(theTag);			// we want to own this one.
			GXLockTag(theTag);
			theData = (curveLayoutRec*)GXGetTagStructure(theTag, nil);

			preserveTransform = GXCloneTransform(GXGetShapeTransform(target) );
			
			if (theData->flags & clSelfContained) {
				GXGetPictureParts(target, 1, 1, &theGlyph, nil, nil, nil);
				GXGetPictureParts(target, 2, 1, &theLayout, nil, nil, nil);
				GXGetPictureParts(target, 3, 1, &thePath, nil, nil, nil);
			} else { 
				theGlyph = target;
				theLayout = theData->theLayout;
				thePath = theData->thePath;
			}//end if
		
			/* Also Rest the layout's Y position to be zero, we always want layout to hug curve, but x value will determine indent */
		
			GXGetLayout(theLayout, nil, nil, nil, nil, nil, nil, nil, nil, &position);
			position.y = ff(0);
			GXSetLayout(theLayout, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, &position);

			(void)MakeTextOnPath(theLayout, thePath, theGlyph, &tooBig);	// pass in theGlyph for old one.
			
			GXSetShapeTags(target, curveLayoutTag, 1, 0, 1, &theTag);			// put the tags back on.
			
			if (tooBig)
				theData->flags |= clTextTooBigForPath;
			else
				theData->flags &= ~clTextTooBigForPath;
			
			GXUnlockTag(theTag);
			GXDisposeTag(theTag);
			
			/* make sure we've preserved the transform */
			
			GXSetShapeTransform(target, preserveTransform);
			GXDisposeTransform(preserveTransform);
			
		} else {
		
			/* if the shape doesn't have our private tag, then it isn't curveLayout */
			
			GXPostGraphicsError(illegal_type_for_shape);

		}//end if
		
	
	}//ClChangedCurveLayout





/*********************************

	Routine ClGetCurveLayoutHighlight
	
	Call this to get highlights instead of GXGetLayoutHighlight.
	
*********************************/
gxShape  ClGetCurveLayoutHighlight(gxShape curveLayout, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight)
	{
		gxShape						text, path;
		gxShape						result;
		gxTransform				theTransform = GXCloneTransform(GXGetShapeTransform(curveLayout));
		curveLayoutFlags	flags;
		
		flags = ClGetCurveLayout(curveLayout, &text, &path);
		
		if (flags & clNonContiguousHighlight)
			result = GetLayoutHighlightOnPath(text, path, startOffset, endOffset, hilightType, highlight);
		else
			result = ContiguousGetLayoutHighlightOnPath(text, path, startOffset, endOffset, hilightType, highlight);

		
		GXSetShapeTransform(result, theTransform);
		GXDisposeTransform(theTransform);

		return(result);
		
	}//ClGetCurveLayoutHighlight





/************************************

	Routine: ClGetCurveLayoutCaret
	
	Call this to get carets instead of GXGetLayoutCaret.


**************************************/
gxShape  ClGetCurveLayoutCaret(gxShape curveLayout, gxByteOffset offset, gxHighlightType highlightType, gxCaretType caretType, gxShape caret)
	{
	
		gxShape						layout, path;
		gxShape						result;
		curveLayoutFlags	flags;
		gxTransform				theTransform = GXCloneTransform(GXGetShapeTransform(curveLayout));
		
		flags = ClGetCurveLayout(curveLayout, &layout, &path);
		
		/********
			If the layout is too big for the path, then ensure we dont' try 
			to get a caret that will be at a character that isn't drawn for that reason.
			This is because layout will give us a caret which when used as a dash
			(that's how we are doing this), it will wrap around to the beginning of the
			curve.
		*********/
		if (flags & clTextTooBigForPath) { 
		
			gxPoint				hitDown;
			gxByteOffset	maxOffset;
		
			/* Make offset they byte offset for the character at the end of the path. */
			
			hitDown.x = ClGetFixedShapeLength(path);		// near, but not at, end of path.
			hitDown.y = ff(0);
			maxOffset = GXHitTestLayout(layout, &hitDown, highlightType, nil, nil);
			if (offset > maxOffset)
				offset = maxOffset;
		
		}//end if
		
		result = GetLayoutCaretOnPath(layout, path, offset, highlightType, caretType, caret);
	
		GXSetShapeTransform(result, theTransform);
		GXDisposeTransform(theTransform);

		return(result);
	
	}//ClGetCurveLayoutCaret




/************************************

	Routine: ClHitTestCurveLayout
	
	Call this to get carets instead of GXHitTestLayout.
	
**************************************/

gxByteOffset ClHitTestCurveLayout(gxShape curveLayout, const gxPoint *hitDown, gxHighlightType highlightType, gxLayoutHitInfo *hitInfo, gxShape hitTrackingArea)
	{
		gxByteOffset		result;
		gxShape					text, path;
		
		ClGetCurveLayout(curveLayout, &text, &path);
		
		result = HitTestLayoutOnPath(GXGetShapeTransform(curveLayout), text, path, hitDown, highlightType, hitInfo, hitTrackingArea);
		
		return(result);
		
	}//ClHitTestCurveLayout




/*****************************************

	Routine to allow application to change
	flags of the curve layout shape.
	
*****************************************/
void ClSetCurveLayoutFlags(gxShape curveLayout, curveLayoutFlags flags);
void ClSetCurveLayoutFlags(gxShape curveLayout, curveLayoutFlags flags)
	{
		gxTag							theTag;
		curveLayoutRec		*theData;
		curveLayoutFlags	oldFlags;
	
		if ( GXGetShapeTags(curveLayout, curveLayoutTag, 1, 1, &theTag) ) {
		
			GXLockTag(theTag);
			theData = (curveLayoutRec*)GXGetTagStructure(theTag, nil);
			
			oldFlags = theData->flags;
			
			/* combine the old non-writable flags with the new writable flags */
			
			theData->flags = (oldFlags & ~clWritableFlagsMask) | ( flags & clWritableFlagsMask );

			GXUnlockTag(theTag);
			
		} else {
		
			GXPostGraphicsError(illegal_type_for_shape);

		}//end if
	
	}//ClSetCurveLayoutFlags


/******************
	ClGetShapeForInk
	
	This routine is called to get a shape reference from the curve layout
	on which to do ink operations. (like GXGetShapeInk, SetShapeInk, GetShapeColor, etc...
	
	This is necessary to allow an implementation of CurveLayout where the CurveLayuot is
	really a picture.  And inks on pictures are ignored.
******************/
gxShape ClGetShapeForInk(gxShape curveLayout)
	{
		long result = ClGetCurveLayout(curveLayout, nil, nil);
		
		if (result & clSelfContained) {	
			gxShape result;
			GXGetPictureParts(curveLayout, 1, 1, &result, nil, nil, nil);
			return(result);
		} else {
			return(curveLayout);
		}//end if
		
	}//ClGetShapeForInk


/******************
	ClShapeIsCurveLayout:
	
	Routine returns true if shape is a curve layout shape,
	otherwise false.
*******************/	
Boolean ClShapeIsCurveLayout(gxShape target);
Boolean ClShapeIsCurveLayout(gxShape target)
	{
	
		if ( GXGetShapeTags(target, curveLayoutTag, 1, 1, nil) )
			return(true);
		else
			return(false);
	
	}//ClShapeIsCurveLayout
