/******************************

	Text on path library Interface file.
	
	Copyright 1996, Apple Computer Inc.  
	
	Daniel I. Lipton
	

	Routines for creating, disposing text-on-path shapes.
	
		1. ClNewCurveLayout:		Make a new shape that is CurveLayout.  This shape can be drawn with GXDrawShape.
		2. ClDisposeCurveLayout		Dispose of the CurveLayout shape.
		3. ClGetCurveLayout			retrieve the original text and path shapes from the CurveLayout shape.
		4. ClChangedCurveLayout		Notify of any change to the text or the path in the CurveLayout shape.
		5. ClSetCurveLayoutFlags	If application wants to change any of the writable flags.
		6. ClGetShapeForInk			Use this to get a shape reference for ink operations.

	These routines have the same API as normal layout routines except they take a shape created with ClMakeTextOnPathShape.

		1. ClGetCurveLayoutHighlight	Get the highlight shape for the CurveLayout.
		2. ClGetCurveLayoutCaret		Get the caret shape for the CurveLayout.
		3. ClHitTestCurveLayout			Hit test CurveLayout.
		
	Handy Utility Routines:
	
		1. ClFindHitLengthAlongPath		Use for getting more info about hit testing on curves.
		2. ClGetShapeForInk				Use to get shape reference of curve layout for ink operations.
											(applying color to curve layout reference directly not guaranteed)
		3. ClShapeIsCurveLayout			Determines if a given shape is a valid curve layout shape.
		
		Notes:
			Layout position is relative to first point of path.
			Layout's transform is ignored.  Apply transformations result of ClNewCurveLayout.

*******************************/


#ifndef TEXTONPATHLIB
#define TEXTONPATHLIB

#include "GXTypes.h"
#include "GXGraphics.h"
#include "GXLayout.h"

enum {

	clNoFlags = 0,
	clTextTooBigForPath = 		0x80000000,			/* Application can show UI if text doesn't fit, flag is read only */
	clSelfContained = 			0x40000000,			/* if curve layout object is self contained/flattenable */
	clNonContiguousHighlight =	0x00000001,			/* fast enough for flicker free with GXExcludeShape, flag is read/write */
	clWritableFlagsMask = 		0x0000FFFF			/* Mask identifying flags that are read/write */
};

typedef long curveLayoutFlags;



/*************
	Notes about the selfContained parameter for ClNewCurveLayout:
	
	If selfContained is true then the curveLayout shape will really be a picture shape.
	The picture will contain the glyph shape that is the curve layout representation.
	It will also contain the original layout and path passed into ClNewCurveLayout.  These
	shapes are added to the picture with an overriding transform that has an empty clip.
	This results in the client being able to draw the picture and only have the glyph
	shape draw, but still store the other two in the picture.
	
	The 2 problems with this approach are as follows:
		1.	Modifiying the Ink of the CurveLayout object directly has no effect since pictures
				ignore inks.
		2.	There is a performance overhead in drawing the picture since it also contains 2 other
				objects in addition to the drawable portion of the CurveLayout object.
		
		The first one is addressed by a utility routine called ClGetShapeForInk and it is used as follows:
			Instead of calling GXSetShapeColor(myCurveLayout),
			  call GXSetShapeColor(ClGetShapeForInk(myCurveLayout));
			
			Do the same for GXSetShapeInk, GXSetShapeTranser.
			
	Originally (selfContained false) the result of ClNewCurveLayout was a glyph shape.
	The glyph shape contained a tag that stored the references to the original layout and
	path shapes.  The problem with this approach was you could't flatten it and then unflatten
	it since the unflattened tag would have meaningless shape references in it.  
	
	The client may choose either method upon creation of the CurveLayout object.
	
	For applications that maintain their own document format it may be best pass true for selfContained
 for performance since the application isn't using flatening anyway.  If the applicaton
	uses the standard GX stream format, it will greatly simplify the code false is passed.
		
	
***************/


gxShape	ClNewCurveLayout(gxShape theText, gxShape theShape, Boolean selfContained);

void  ClDisposeCurveLayout(gxShape curveLayout);

curveLayoutFlags ClGetCurveLayout(gxShape curveLayout, gxShape *theText, gxShape *thePath);  // returns flags.

void ClChangedCurveLayout(gxShape curveLayout);

void ClSetCurveLayoutFlags(gxShape curveLayout, curveLayoutFlags flags);

gxShape  ClGetCurveLayoutHighlight(gxShape curveLayout, gxByteOffset startOffset, gxByteOffset endOffset, gxHighlightType hilightType, gxShape highlight);

gxShape  ClGetCurveLayoutCaret(gxShape curveLayout, gxByteOffset offset, gxHighlightType highlightType, gxCaretType caretType, gxShape caret);

gxByteOffset ClHitTestCurveLayout(gxShape curveLayout, const gxPoint *hitDown, gxHighlightType highlightType, gxLayoutHitInfo *hitInfo, gxShape hitTrackingArea);

Fixed ClFindHitLengthAlongPath(gxShape path, const gxPoint *hitPoint, gxPoint *pointOnCurve);

gxShape ClGetShapeForInk(gxShape curveLayout);

Boolean ClShapeIsCurveLayout(gxShape target);

#endif

